#include "constants.h"
#include "TBoard.h"

TBoard::TBoard(WINDOW *_W_mi_campo, WINDOW *_W_campo_enemigo, WINDOW *_W_comandos, int y, int x, int _player) {

  W_mi_campo = _W_mi_campo;
  W_campo_enemigo = _W_campo_enemigo;
  W_comandos = _W_comandos;
  continuar = true;

  parent_x = x;
  parent_y = y;
  player = _player;

  //Dibuja el "board" (tablero) de juego en mi campo y en el campo enemigo
  draw_board(Y_BOARD, X_BOARD, W_mi_campo);
  draw_board(Y_BOARD, X_BOARD, W_campo_enemigo);

  //dependiendo de que jugador inicie el juego se debe crear o leer el espacio de memoria compartido
  if (player == 1) {
    create_shared_memory();
    campo_mio = buf;
    campo_enemigo = buf_sp;
  }
  else {
    read_shared_memory();
    campo_mio = buf_sp;
    campo_enemigo = buf;
  }

  /*refresca las ventanas para que se vea el tablero (ver documentacion de
   *ncurses* acerca del funcionamiento de la instruccion refresh o wrefresh)*/
  wrefresh(W_mi_campo); wrefresh(W_campo_enemigo);
}

/*
Funcion para crear la memoria compartida
*/
void TBoard::create_shared_memory(){
  int i;

  //crea bufer con arreglo de la pantalla de la izquierda
  for(i=0;i<100;i++){
    if( (shmid[i] = shmget(SHMKEY +i,sizeof(BUFFER), PERMS | IPC_CREAT)) < 0)
      printf("server: can't get shared memory %d",i);
    if((buf[i] = (BUFFER *) shmat(shmid[i], 0,0))==(BUFFER *) -1)
      printf("server:can't attach shared memory %d",i); 
     //llenar buf con 0's
    buf[i]->data=0;
  }

  //crea memoria para la pantalla de la derecha
  for(i=0;i<100;i++){
    // SHMKEY+i+120 el 1200 es porque las direcciones anteriores ya fueron ocupadas al crear el 1er espacio compartido
    if((shmid_sp[i] = shmget(SHMKEY+i+120,sizeof(BUFFER),PERMS | IPC_CREAT)) < 0)
      printf("server: can't get shared memory %d",i);
    if((buf_sp[i]=(BUFFER *) shmat(shmid_sp[i],0,0))==(BUFFER *) -1)
      printf("server: can't attach shared memory %d",i);
    buf_sp[i]->data=0;
  }

  //crea memoria compartida para pseudo_semaforo para poder sincronizar turnos
  if((shmid_turno = shmget(SHMKEY+250,sizeof(BUFFER),PERMS | IPC_CREAT)) < 0)
    printf("server: can't get shared memory turn");
  if((buf_turno[0]=(BUFFER *) shmat(shmid_turno,0,0))==(BUFFER *) -1)
    printf("server: can't attach shared memory turn");
  
  //player 1 es el primero en jugar, "2" para jugador 2
  buf_turno[0]->data=1;
}

/*
Lectura de memoria compartida
*/
void TBoard::read_shared_memory(){
   int i;
   
   //se lee la memoria compartida de los dos buffer creados anteriormente
   for(i=0;i<100;i++){
  //primer buffer
  if((shmid[i] = shmget(SHMKEY+i,sizeof(BUFFER),PERMS | IPC_CREAT)) < 0)
      printf("server: can't get shared memory %d",i);
  if((buf[i]=(BUFFER *) shmat(shmid[i],0,0))==(BUFFER *) -1)
      printf("server: can't attach shared memory %d",i);
  
  //segundo buffer   
  if((shmid_sp[i]=shmget(SHMKEY+i+120,sizeof(BUFFER),PERMS | IPC_CREAT))<0)
      printf("server: can't get shared memory %d",i);
  if((buf_sp[i]=(BUFFER *) shmat(shmid_sp[i],0,0))==(BUFFER *) -1)
      printf("server: can't attach shared memory %d",i);
   }

   //se lee ademas el la variable para conocer la persona que esta de turno
   if((shmid_turno = shmget(SHMKEY+250,sizeof(BUFFER),PERMS | IPC_CREAT)) < 0)
     printf("server: can't get shared memory turn");
   if((buf_turno[0]=(BUFFER *) shmat(shmid_turno,0,0))==(BUFFER *) -1)
     printf("server: can't attach shared memory turn");

}

/*
Inicio del juego, incluye bienvenida y se pide la insercion de botes
*/
void TBoard::start(int debug_level) {
  if (player == 1)
    log("Bienvenido jugador 1 presione <enter> para crear el juego");
  else
    log("Bienvenido jugador 2 presione <enter> para conectarse al juego");
  wgetch(W_comandos);

  if (debug_level == 0) {
    insert_boats(3); //XXX cambiar para preguntar la cantidad de barcos que va a colocar
  } else if (debug_level == 1) {
    insert_boats_auto();
  }
}
bool TBoard::revisar_si_perdi() {
  int perdi = 100;
  for (unsigned int i=0; i<100; i++) {
    perdi--;
    if (campo_mio[i]->data == 1)
      break;
  }
  if (perdi == 0)
    return true;
  else
    return false;
}

bool TBoard::revisar_si_gane(){
  int gane = 100;
  for (unsigned int i=0; i<100; i++) {
    gane--;
    if (campo_enemigo[i]->data == 1)
      break;
  }
  if (gane == 0)
    return true;
  else
    return false;
}

/*
Dibujar el tablero de juego
*/
void TBoard::draw_board(int y, int x, WINDOW *ventana) {
  mvwprintw(ventana, y,    x,  "10| | | | | | | | | | |");
  mvwprintw(ventana, y+1,  x,  " 9| | | | | | | | | | |");
  mvwprintw(ventana, y+2,  x,  " 8| | | | | | | | | | |");
  mvwprintw(ventana, y+3,  x,  " 7| | | | | | | | | | |");
  mvwprintw(ventana, y+4,  x,  " 6| | | | | | | | | | |");
  mvwprintw(ventana, y+5,  x,  " 5| | | | | | | | | | |");
  mvwprintw(ventana, y+6,  x,  " 4| | | | | | | | | | |");
  mvwprintw(ventana, y+7,  x,  " 3| | | | | | | | | | |");
  mvwprintw(ventana, y+8,  x,  " 2| | | | | | | | | | |");
  mvwprintw(ventana, y+9,  x,  " 1| | | | | | | | | | |");
  mvwprintw(ventana, y+10, x,  "  |A|B|C|D|E|F|G|H|I|J|");
}

/*
En el tablero se le piden letras al usuarios ya que asi esta diseñado el juego,
esta funcion mapea esas letras a numeros para poder ubicar mas facilmente los barcos.
*/
int TBoard::letra_a_entero(char user_input){
  int user_int;
  //dependiendo de la letra entregada se pasa a numero para mostrarla en el grafico
  if(user_input == 'A' || user_input == 'a') return user_int = 1;
  if(user_input == 'B' || user_input == 'b') return user_int = 2;
  if(user_input == 'C' || user_input == 'c') return user_int = 3;
  if(user_input == 'D' || user_input == 'd') return user_int = 4;
  if(user_input == 'E' || user_input == 'e') return user_int = 5;
  if(user_input == 'F' || user_input == 'f') return user_int = 6;
  if(user_input == 'G' || user_input == 'g') return user_int = 7;
  if(user_input == 'H' || user_input == 'h') return user_int = 8;
  if(user_input == 'I' || user_input == 'i') return user_int = 9;
  if(user_input == 'J' || user_input == 'j') return user_int = 10;
  //retorna 0 si esta fuera del rango
  return 0;
}

/*
Funcion para pedirle al usuario un numero entero. La funcion escribe el texto,
coloca un offset, y lo que el usuario coloca se devuelve como entero
*/
int TBoard::ask_integer(int offset, const char *texto) {
  char user_input[10];

  mvwprintw(W_comandos, offset, X_BOARD, "%s", texto);
  wrefresh(W_comandos);
  wgetstr(W_comandos, user_input);

  return atoi(user_input);
}

/*
Escribe el texto en pantalla, con un offset vertical entregados ambos como
parametros. lo que ingresa el usuario se devuelve casteado a char
*/
char TBoard::ask_char(int offset, const char *texto) {
  int user_input;
  mvwprintw(W_comandos, offset, X_BOARD, "%s", texto);

  user_input = wgetch(W_comandos);
  return (char) user_input;
}

/*
Se actualiza la coordenada deseada en el sector de memoria compartida
*/
void TBoard::coloca_coordenada_mem(int user_y, int user_x, int modo) {
   campo_mio[user_y*10 + user_x]->data = modo;
}

/*
Dibuja en los tableros la ubicacion de los elementos 
*/
void TBoard::pinta_punto(int user_y, int user_x, int tipo) {
  wmove(W_mi_campo, Y_BOARD + 10 - (user_y), X_BOARD + (user_x * 2) + 1);
  if (tipo == AGUA)
    waddch(W_mi_campo, ' ' | A_BOLD);
  else if (tipo == BARCO) 
    waddch(W_mi_campo, 'A' | A_BOLD);
  else if (tipo == FALLADO) 
    waddch(W_mi_campo, '*' | A_BOLD);
  else if (tipo == ACERTADO)
    waddch(W_mi_campo, 'X' | A_BOLD);
}
void TBoard::pinta_punto(int user_y, int user_x, int tipo, WINDOW *ventana) {
  wmove(ventana, Y_BOARD + 10 - (user_y), X_BOARD + (user_x * 2) + 1);
  if (tipo == AGUA)
    waddch(ventana, ' ' | A_BOLD);
  else if (tipo == BARCO) 
    waddch(ventana, 'A' | A_BOLD);
  else if (tipo == FALLADO) 
    waddch(ventana, '*' | A_BOLD);
  else if (tipo == ACERTADO)
    waddch(ventana, 'X' | A_BOLD);
  else {
    log("error al pintar punto, presione <enter>");
    wgetch(W_comandos);
  }
}

/*
Se entrega una posicion del tablero izquierdo, luego se dibuja en este un caracter entregado que 
puede representar barco, agua, golpe en barco o golpe en agua
*/
void TBoard::refresca_desde_memoria(int i, WINDOW *ventana_destino, char dibujo) {
  wattron(ventana_destino, COLOR_PAIR(1));
  wmove(ventana_destino, Y_BOARD + 10 - (i / 10), X_BOARD + ((i%10) * 2) + 1);
  waddch(ventana_destino, dibujo | A_BOLD);
  wattroff(ventana_destino, COLOR_PAIR(1));
}

/*

*/
void TBoard::refresca_coordenada_enemiga(int index) {
  if(buf_sp[index]->data==1)
    refresca_desde_memoria(index, W_campo_enemigo,'A');
}
/*
Se entrega una posicion del tablero, si en esta hay un barco se cambia el dato por un * (barco alcanzado)
y si no se marca como golpe al agua
*/
void TBoard::refresca_coordenada_mia(int index) {
 if (campo_mio[index]->data == ACERTADO)
     refresca_desde_memoria(index, W_mi_campo,'*');
  else if(campo_mio[index]->data == FALLADO)
     refresca_desde_memoria(index, W_mi_campo,'0');
}

/*
Actualiza por separado cada posicion de ambos tableros
*/
void TBoard::actualizar_todos_tableros(){
  int i;
  //jugadores leen la pantalla del otro 
  for(i=0;i<100;i++){
    refresca_coordenada_mia(i);
    refresca_coordenada_enemiga(i);
  }
  wrefresh(W_mi_campo);
  wrefresh(W_campo_enemigo);
}

/*
El jugador que lo ejecuta lee la pantalla del otro para actualizar los elementos en pantalla
*/
void TBoard::actualiza_mi_tablero() {
  for(int i=0;i<100;i++){
    refresca_coordenada_mia(i);
  }
  wrefresh(W_mi_campo);
}

/*
Dibuja un barquito con letras "A", basandose en las coodenadas user_x y user_x
junto a la orientacion que se entrega como argumentos de la funcion.
*/
void TBoard::dibuja_barco(char orientacion, int user_y, int user_x) {
  if (orientacion == 'N' || orientacion == 'n') {
    for (unsigned int i=0; i < TAMANIO_BARCO; i++) {
      pinta_punto(user_y + i, user_x, 1);
      //llena el buffer con las coordenadas del barco
      coloca_coordenada_mem(user_y+i, user_x, 1);
    }
  }
  if (orientacion == 'E' || orientacion == 'e') {
    for (unsigned int i=0; i < TAMANIO_BARCO; i++) {
      pinta_punto(user_y, user_x + i, 1);
      coloca_coordenada_mem(user_y, user_x+i, 1);
    }
  }
}

/*
Se entrega una posicion en la cual se chequeara si hay o no un barco enemigo, dependiendo de la
resolucion se actualizara el tablero 
*/
void TBoard::fuego(int user_y, int user_x) {
  int  index = user_y*10 + user_x;

  wattron(W_campo_enemigo, COLOR_PAIR(1));  
  if(campo_enemigo[index]->data == AGUA){
  campo_enemigo[index]->data = FALLADO;
  pinta_punto(user_y,user_x,FALLADO,W_campo_enemigo);
  }
  else if(campo_enemigo[index]->data == BARCO){
  campo_enemigo[index]->data = ACERTADO;
  pinta_punto(user_y,user_x,ACERTADO,W_campo_enemigo);
  }

  wattroff(W_campo_enemigo, COLOR_PAIR(1));
  wrefresh(W_campo_enemigo);
}

/*
 * Este metodo se ocupa en el modo manual (DEBUG == 0). A diferencia de la del
 * modo automatico, este metodo no ocupa argumentos de entrada, y se le pide al
 * usuario en cambio que ingrese las coordenadas en donde desea disparar
 */
void TBoard::turno() {
  int user_x, user_y;
  char comodin;

  log("Esperando al otro jugador");
  while(buf_turno[0]->data != player);

  if (revisar_si_perdi()) {
    log("PERDISTE!");
    wgetch(W_comandos);
    continuar = false;
  }
  actualiza_mi_tablero();
  werase(W_comandos);
  wattron(W_comandos, COLOR_PAIR(1));
  comodin = ask_char(0, "Ingresa coordenada X del disparo? ");
  user_y = ask_integer(1, "Ingresa coordenada Y del disparo? ");
  user_x = letra_a_entero(comodin);
  wattron(W_comandos, COLOR_PAIR(1));

  fuego(user_y, user_x);
  wrefresh(W_campo_enemigo);
  if (revisar_si_gane()) {
    log("BIEN, GANASTE!");
    wgetch(W_comandos);
    continuar = false;
  }
  if(player == 1)
    buf_turno[0]->data=2;
  if(player == 2)
    buf_turno[0]->data=1;
}

/*
 * Este se ocupa para hacer los turnos automaticos (DEBUG == 1), se le entrega como argumento
 * de entrada las coordenadas y e x en donde el computador va a disparar
 */
void TBoard::turno(int user_y, int user_x) {
  log("Esperando al otro jugador");
  while(buf_turno[0]->data != player);

  if (revisar_si_perdi()) {
    log("PERDISTE!");
    wgetch(W_comandos);
    continuar = false;
  }
  log("FUEGO!!!");
  wgetch(W_comandos);
  actualiza_mi_tablero();
  fuego(user_y, user_x);
  wrefresh(W_campo_enemigo);
  if (revisar_si_gane()) {
    log("BIEN, GANASTE!");
    wgetch(W_comandos);
    continuar = false;
  }
  if(player == 1)
    buf_turno[0]->data=2;
  if(player == 2)
    buf_turno[0]->data=1;
}

/*
Se solicitan posiciones para colocar los barcos en el tablero.
*/
void TBoard::insert_boats(unsigned int n) {
  int user_x, user_y;
  char comodin;
  char orientacion;

  for (unsigned int i=0; i<n; i++) {
    werase(W_comandos);

    // Pide las coordenadas X e Y del barco a colocar, pide ademas la orientacion
    comodin = ask_char(0, "Ingresa coordenada X del primer barco ? ");
    user_y = ask_integer(1, "Ingresa coordenada Y del primer barco ? ");
    user_x = letra_a_entero(comodin);
    orientacion = ask_char(2, "Ingresa orientacion del barco (N: Norte, E: Este)");
    mvprintw(Y_INPUT + 3, X_BOARD, "Ingreso: %c", orientacion);

    /*activa el atributo de color rojo para entre medio (dentro del if, si
    orientacion es una orientacion valida) dibujar el barco con la orientacion
    y las coordenadas que el usuario acaba de entregar*/
    wattron(W_mi_campo, COLOR_PAIR(1));
    if (orientacion == 'N' || orientacion == 'n' || orientacion == 'E' || orientacion == 'e')
      dibuja_barco(orientacion, user_y, user_x);
    wattroff(W_mi_campo, COLOR_PAIR(1));
    wrefresh(W_mi_campo);
  }
}

/*
--------------------------- Metodos para realizar debug ---------------------------------
*/
/*
Realiza automaticamente la insercion de botes para acelerar la prueba del juego 
*/
void TBoard::insert_boats_auto() {
  int user_x, user_y;

  for (unsigned int i=0; i<3; i++) {
    werase(W_comandos);

    // Pide las coordenadas X e Y del barco a colocar, pide ademas la orientacion
    if (player == 1) {
      user_y = (i+1) * 2;
      user_x = (i+1) * 2;
    }
    else {
      user_y = (7-(i*2));
      user_x = (7-(i*2));
    }

    //activa el atributo de color rojo para entre medio (dentro del if, si
    //orientacion es una orientacion valida) dibujar el barco con la orientacion
    //y las coordenadas que el usuario acaba de entregar
    wattron(W_mi_campo, COLOR_PAIR(3));
    if ((i % 2) == 0)
      dibuja_barco('E', user_y, user_x);
    else
      dibuja_barco('N', user_y, user_x);
    wattroff(W_mi_campo, COLOR_PAIR(3));
    wrefresh(W_mi_campo);
  }
}

/*
Muesta mensajes de log para que el usuario pueda tener feedback de lo que esta sucediendo
*/
void TBoard::log(const char *texto) {
  werase(W_comandos);
  wattron(W_comandos, COLOR_PAIR(2));
  mvwprintw(W_comandos, TAMANIO_COMANDOS-1, X_BOARD, texto);
  wattroff(W_comandos, COLOR_PAIR(2));
  wrefresh(W_comandos);
}
void TBoard::log(const char *texto, int numero) {
  werase(W_comandos);
  wattron(W_comandos, COLOR_PAIR(2));
  mvwprintw(W_comandos, TAMANIO_COMANDOS-1, X_BOARD, texto, numero);
  wattroff(W_comandos, COLOR_PAIR(2));
  wrefresh(W_comandos);
}
